# ----------------------------------------------------------------------
# Author:        yury.matveev@desy.de
# ----------------------------------------------------------------------

"""Axis camera proxy
"""

import time
import numpy as np
import logging

try:
    import PyTango
except ImportError:
    pass

from distutils.util import strtobool

from petra_camera.devices.base_camera import BaseCamera
from petra_camera.main_window import APP_NAME

logger = logging.getLogger(APP_NAME)

_base_settings_map = {'FPS': ("device_proxy", "FPSLimit"),
                      'max_width': ("device_proxy", "ImageWidth"),
                      'max_height': ("device_proxy", "ImageHeight"),
                      'pan': ("device_proxy", "Pan"),
                      'max_pan': ("device_proxy", "MaxPan"),
                      'min_pan': ("device_proxy", "MinPan"),
                      'tilt': ("device_proxy", "Tilt"),
                      'max_tilt': ("device_proxy", "MaxTilt"),
                      'min_tilt': ("device_proxy", "MinTilt"),
                      'zoom': ("device_proxy", "Zoom"),
                      'max_zoom': ("device_proxy", "MaxZoom"),
                      'min_zoom': ("device_proxy", "MinZoom"),
                      'focus': ("device_proxy", "Focus"),
                      'max_focus': ("device_proxy", "MaxFocus"),
                      'min_focus': ("device_proxy", "MinFocus"),
                      }


# ----------------------------------------------------------------------
class AXISCamera(BaseCamera):
    """Proxy to a physical TANGO device.
    """

    START_DELAY = 1
    STOP_DELAY = 0.5

    START_ATTEMPTS = 2

    visible_layouts = ('FPS', 'position')

    # ----------------------------------------------------------------------
    def __init__(self, settings):
        self._settings_map = dict(_base_settings_map)
        super(AXISCamera, self).__init__(settings)

        if 'color' in settings.keys():
            self.color = strtobool(settings.get("color"))
        else:
            self.color = False

        self._image_source = 'imagergb' if self.color else 'imagegrayscale'

        self._last_frame = np.zeros((1, 1))
        self._last_time = time.time()

        self._eid = None

        self._main_runner = False

    # ----------------------------------------------------------------------
    def get_settings(self, option, cast, do_rotate=True, do_log=True):

        if option == 'FPSmax':
            logger.debug(f'{self._my_name}: setting {cast.__name__}({option}) requested')
            return 200
        elif option in ['pan', 'tilt', 'zoom', 'focus']:
            return super(AXISCamera, self).get_settings(option, cast, do_rotate, False)
        else:
            return super(AXISCamera, self).get_settings(option, cast, do_rotate, do_log)

    # ----------------------------------------------------------------------
    def save_settings(self, option, value):

        if option == 'FPS':
            need_restart = False
            if self._device_proxy.state() == PyTango.DevState.RUNNING:
                need_restart = True
                self._device_proxy.command_inout("StopAcquisition")
                while self._device_proxy.state() == PyTango.DevState.RUNNING:
                    time.sleep(0.1)
            super(AXISCamera, self).save_settings(option, value)
            if need_restart:
                self._device_proxy.command_inout("StartAcquisition")
        else:
            super(AXISCamera, self).save_settings(option, value)

    # ----------------------------------------------------------------------
    def _start_acquisition(self):
        """
        """

        if self._device_proxy.state() in [PyTango.DevState.ON, PyTango.DevState.RUNNING]:

            logger.debug(f'{self._my_name}: starting acquisition')

            self._mode = 'event'
            self._eid = self._device_proxy.subscribe_event(self._image_source,
                                                           PyTango.EventType.CHANGE_EVENT,
                                                           self._readout_frame)

            if self._device_proxy.state() == PyTango.DevState.ON:
                self._main_runner = True
                attemp = 0
                while attemp < self.START_ATTEMPTS:
                    try:
                        self._device_proxy.command_inout("StartAcquisition")
                        break
                    except:
                        attemp += 1
            else:
                self._main_runner = False

            time.sleep(self.START_DELAY)

            return True

        else:
            logger.warning("Camera should be in ON state (is it running already?)")
            return False

    # ----------------------------------------------------------------------
    def stop_acquisition(self):
        """
        """
        if self._eid is not None:
            self._device_proxy.unsubscribe_event(self._eid)
            self._eid = None
        if self._main_runner:
            self._device_proxy.command_inout("StopAcquisition")

        time.sleep(self.STOP_DELAY)  # ? TODO

    # ----------------------------------------------------------------------
    def is_running(self):
        return self._device_proxy.state() == PyTango.DevState.RUNNING and self._eid is not None

    # ----------------------------------------------------------------------
    def _readout_frame(self, event):
        """Called each time new frame is available.
        """
        if not event.err:
            try:
                data = event.attr_value
                if data.value is not None:
                    self._last_frame = self._process_frame(data.value)
                    self._new_frame_flag = True
            except Exception as err:
                logger.error('AXISCamera error: {}'.format(err))
                self.error_flag = True
                self.error_msg = str(err)
        else:
            logger.error('AXISCamera error: {}'.format(self.error_msg))
            self.error_flag = True
            self.error_msg = event.errors

    # ----------------------------------------------------------------------
    def _process_frame(self, data):
        data = np.transpose(data)
        if self.color:
            c_data = np.zeros(data.shape + (3,), dtype=np.ubyte)
            c_data[..., 0] = data & 255
            c_data[..., 1] = (data >> 8) & 255
            c_data[..., 2] = (data >> 16) & 255

            return c_data
        else:
            return data

    # ----------------------------------------------------------------------
    def close_camera(self):
        super(AXISCamera, self).close_camera()

        if self._eid is not None:
            self._device_proxy.unsubscribe_event(self._eid)



# copy of necessary HasyUtils functions in case there is not HasyUtils installation
import sys
import PyTango

_db = None

def getDeviceNamesByClass(className, tangoHost=None):
    '''Return a list of all devices of a specified class,
        'DGG2' -> ['p09/dgg2/exp.01', 'p09/dgg2/exp.02']
    '''
    srvs = getServerNameByClass(className, tangoHost)
    argout = []

    db = _findDB(tangoHost)
    if not db:
        return None

    for srv in srvs:
        lst = db.get_device_name(srv, className).value_string
        for i in range(0, len(lst)):
            argout.append(lst[i])
    return argout

def getServerNameByClass(argin, tangoHost = None):
    '''Return a list of servers containing the specified class '''

    db = _findDB( tangoHost)

    srvs = db.get_server_list( "*").value_string

    argout = []

    for srv in srvs:
        classList = db.get_server_class_list( srv).value_string
        for clss in classList:
            if clss == argin:
                argout.append( srv)
                break
    return argout

def _findDB(tangoHost = None):
    '''
    handle these cases:
      - tangoHost == None: use TANGO_HOST DB
      - tangoHost == "haspp99:10000" return db link
      - tangoHost == "haspp99" insert 100000 and return db link
    '''
    if tangoHost is None:
        if _db is None:
            print( "TgUtils._findDB: _db is None")
            sys.exit( 255)
        return _db

    #
    # unexpeccted: tango://haspe212oh.desy.de:10000/motor/dummy_mot_ctrl/1
    #
    if tangoHost.find( 'tango://') == 0:
        print( "TgUtils._fineDB: bad TANGO_HOST syntax %s" % tangoHost)
        return None
    #
    # tangHost "haspp99:10000"
    #
    lst = tangoHost.split( ':')
    if len(lst) == 2:
        return PyTango.Database(lst[0], lst[1])
    #
    # tangHost "haspp99"
    #
    elif len(lst) == 1:
        return PyTango.Database(lst[0], "10000")
    else:
        print("TgUtils._fineDB: failed to return Database, %s" % tangoHost)
        return None


# ----------------------------------------------------------------------
# Author:        yury.matveev@desy.de
# ----------------------------------------------------------------------

"""Dummy 2D data generator.
"""
import threading
import time
import requests
from random import random
import io
from PIL import Image
import numpy as np
import logging
import PyTango

from threading import Thread

from petra_camera.devices.base_camera import BaseCamera
from petra_camera.main_window import APP_NAME

logger = logging.getLogger(APP_NAME)

DEFAULT_TANGO_SERVER = 'hasep23swt01:10000/pxx/petrastatusscreen/all'


# ----------------------------------------------------------------------
class PetraStatus(BaseCamera):
    """
    """
    _settings_map = {'max_width': ('self', 'FRAME_W'),
                     'max_height': ('self', 'FRAME_H')}

    visible_layouts = ('FPS')

    FRAME_W = 1122
    FRAME_H = 898

    # ----------------------------------------------------------------------
    def __init__(self, settings):
        super(PetraStatus, self).__init__(settings)

        self._fps = self.get_settings('FPS', int)
        if self._fps == 0:
            self._fps = 1

        if 'status_source' in settings.keys():
            self._mode = settings.get("status_source")
        else:
            self._mode = 'tango'

        if self._mode == 'tango' and 'server' in settings.keys():
            self._tango_server = settings.get("server")
        else:
            self._tango_server = DEFAULT_TANGO_SERVER

        if self._mode != 'tango':
            self.FRAME_W = 800
            self.FRAME_H = 600

        self._picture_thread = Thread(target=self._readout_frame)
        self._generator_thread_working = False
        self._stop_picture_thread = threading.Event()
        self._run_acquisition = threading.Event()
        self._picture_thread.start()

    # ----------------------------------------------------------------------
    def close_camera(self):

        self._stop_picture_thread.set()
        while self._generator_thread_working:
            time.sleep(1/self._fps)

    # ----------------------------------------------------------------------
    def _readout_frame(self):

        while not self._stop_picture_thread.is_set():

            time.sleep(1 / self._fps)

            if self._run_acquisition.is_set():
                try:
                    if self._mode == 'tango':
                        data = PyTango.DeviceProxy(self._tango_server).statusscreen[
                               self._picture_size[0]:self._picture_size[2], self._picture_size[1]:self._picture_size[3]]
                        c_data = np.zeros(data.shape + (3,), dtype=np.ubyte)
                        c_data[..., 0] = data & 255
                        c_data[..., 1] = (data >> 8) & 255
                        c_data[..., 2] = (data >> 16) & 255
                        self._last_frame = c_data
                        self._last_camera_msg = PyTango.DeviceProxy(self._tango_server).laststatusmessage
                        self._new_frame_flag = True
                        self._new_msg_flag = True
                        logger.debug(f"{self._my_name} new frame")
                    else:
                        ans = requests.get(f'https://winweb.desy.de/mca/accstatus/infoscreen/petra_status_800.png?{random()}')
                        if ans.status_code == 200:
                            picture_stream = io.BytesIO(ans.content)
                            picture = Image.open(picture_stream)
                            frame = np.rot90(np.asarray(picture, dtype=np.int32), 1)[::-1, :]
                            self._last_frame = frame[self._picture_size[0]: self._picture_size[2],
                                                     self._picture_size[1]: self._picture_size[3]]
                            self._new_frame_flag = True
                            logger.debug(f"{self._my_name} new frame")
                except Exception as err:
                    logger.error(f"{self._my_name}: cannot get new frame: {repr(err)}")

        self._generator_thread_working = False

    # ----------------------------------------------------------------------
    def _start_acquisition(self):

        logger.debug(f"{self._my_name} starting thread")

        self._run_acquisition.set()
        return True

    # ----------------------------------------------------------------------
    def stop_acquisition(self):
        self._run_acquisition.clear()

    # ----------------------------------------------------------------------
    def is_running(self):

        return self._run_acquisition.is_set()

    # ----------------------------------------------------------------------
    def get_settings(self, option, cast, do_rotate=True, do_log=True):

        if option in ['FPSmax', 'max_width', 'max_height']:

            logger.debug(f'{self._my_name}: setting {cast.__name__}({option}) requested')

            if option == 'FPSmax':
                return 200
            elif option == 'max_width':
                return self.FRAME_W
            elif option == 'max_height':
                return self.FRAME_H
        else:
            return super(PetraStatus, self).get_settings(option, cast, do_rotate, do_log)

    # ----------------------------------------------------------------------
    def save_settings(self, option, value):

        if option == 'FPS':

            logger.debug(f'{self._my_name}: setting {option}: new value {value}')
            self._fps = value

        super(PetraStatus, self).save_settings(option, value)